import json
import os
from unittest import mock

from django.contrib.staticfiles.finders import find
from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.test import RequestFactory, TestCase
from django.test.utils import override_settings
from django.urls import reverse
from django.utils.http import urlencode

from galleryfield import defaults
from galleryfield import image_views as built_in_views
from galleryfield.models import BuiltInGalleryImage
from tests import factories
from tests.mixins import UserCreateMixin
from tests.utils import (get_upload_file_path, remove_upload_directory,
                         test_media_root)


class ViewTestMixin(UserCreateMixin):
    @classmethod
    def setUpTestData(cls):  # noqa
        super().setUpTestData()
        cls.user = cls.create_user()

    def setUp(self) -> None:
        factories.UserFactory.reset_sequence()
        factories.BuiltInGalleryImageFactory.reset_sequence()
        factories.DemoGalleryFactory.reset_sequence()
        self.factory = RequestFactory()

        self.c.logout()
        super().setUp()
        remove_upload_directory()

    @staticmethod
    def get_demo_form_view_url():
        return reverse("gallery")

    @staticmethod
    def get_demo_upload_url():
        return reverse(defaults.DEFAULT_UPLOAD_URL_NAME)

    @staticmethod
    def get_demo_fetch_url(params=None):
        url = reverse(defaults.DEFAULT_FETCH_URL_NAME)
        params = params or {}
        if params:
            url += "?" + urlencode(params)
        return url

    @staticmethod
    def get_demo_crop_url(pk):
        return reverse(defaults.DEFAULT_CROP_URL_NAME, kwargs={"pk": pk})

    @staticmethod
    def _get_crop_post_data(
            cropped_result=None,
            thumbnail_size=defaults.DEFAULT_THUMBNAIL_SIZE):
        data = {}
        if cropped_result:
            data.update({"cropped_result": cropped_result})
        if thumbnail_size:
            data.update({"thumbnail_size": thumbnail_size})

        return data

    @staticmethod
    def get_cropped_result(**kwargs):
        data = {}
        for v in ["x", "y", "width", "height", "rotate"]:
            value = kwargs.get(v)
            if value is not None:
                data[v] = value

        return json.dumps(data)

    def demo_upload_file(self, user, file_obj, force_login=True, xml_request=True):
        if force_login:
            self.c.force_login(user)
        with open(file_obj, "rb") as fp:
            request_kwargs = {}
            if xml_request:
                request_kwargs["HTTP_X_REQUESTED_WITH"] = "XMLHttpRequest"
            resp = self.c.post(
                self.get_demo_upload_url(),
                data={"files[]": fp}, **request_kwargs)
        return resp

    def get_demo_crop_pk(self, idx):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5, shuffle=True)
        pk = gallery.images[idx]
        return pk

    def get_default_crop_post_data(self, **kwargs):
        data = self.get_crop_post_data(**kwargs)
        return data


@override_settings(MEDIA_ROOT=test_media_root)
class GalleryWidgetUploadViewTest(ViewTestMixin, TestCase):
    def test_get_gallery_page(self):
        self.c.force_login(self.user)
        resp = self.c.get(self.get_demo_form_view_url())
        self.assertEqual(resp.status_code, 200)

    def test_upload_invalid_file_type(self):
        self.c.force_login(self.user)
        file = get_upload_file_path("test_file.pdf")

        with open(file, "rb") as fp:
            resp = self.c.post(
                self.get_demo_upload_url(),
                data={"files[]": fp}, HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 400)

        expected_error = ("Upload a valid image. The file you uploaded "
                          "was either not an image or a corrupted image.")
        errors = json.loads(resp.content)["errors"]["image"]
        self.assertIn(expected_error, errors)
        self.assertEqual(BuiltInGalleryImage.objects.count(), 0)

    def test_upload_success(self):
        resp = self.demo_upload_file(self.user, find("demo/screen_upload.png"))
        self.assertEqual(resp.status_code, 200, resp.content)
        self.assertEqual(BuiltInGalleryImage.objects.count(), 1)

    def test_upload_fail_not_xml_request(self):
        resp = self.demo_upload_file(self.user, find("demo/screen_upload.png"),
                                     xml_request=False)
        self.assertEqual(resp.status_code, 403, resp.content)
        self.assertEqual(BuiltInGalleryImage.objects.count(), 0)

    def test_upload_not_logged_in(self):
        resp = self.demo_upload_file(self.user, find("demo/screen_upload.png"),
                                     force_login=False)
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(BuiltInGalleryImage.objects.count(), 0)

    def test_cbv_target_model_not_configured(self):
        file = get_upload_file_path("test_file.pdf")

        with open(file, "rb") as fp:
            data = {"files[]": fp}
            request = self.factory.post(
                self.get_demo_upload_url(), data=data,
                HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageCreateView.as_view(
                target_model=None)(request)

        expected_msg = ("Using BaseImageModelMixin (base class of "
                        "BuiltInImageCreateView) "
                        "without the 'target_model' attribute is prohibited.")
        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_disable_server_side_crop(self):
        file = find("demo/screen_upload.png")

        with open(file, "rb") as fp:
            data = {"files[]": fp}
            request = self.factory.post(
                self.get_demo_upload_url(), data=data,
                HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        resp = built_in_views.BuiltInImageCreateView.as_view(
            disable_server_side_crop=True)(request)

        self.assertEqual(resp.status_code, 200, resp.content)

    def test_cbv_invalid_crop_url(self):
        file = find("demo/screen_upload.png")

        with open(file, "rb") as fp:
            data = {"files[]": fp}
            request = self.factory.post(
                self.get_demo_upload_url(), data=data,
                HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageCreateView.as_view(
                crop_url_name="invalid-crop-url")(request)

        expected_msg = ("'crop_url_name' in BuiltInImageCreateView "
                        "is invalid. The exception is: NoReverseMatch: "
                        "Reverse for 'invalid-crop-url' not found.")

        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_crop_url_not_configured(self):
        file = find("demo/screen_upload.png")

        with open(file, "rb") as fp:
            data = {"files[]": fp}
            request = self.factory.post(
                self.get_demo_upload_url(), data=data,
                HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        resp = built_in_views.BuiltInImageCreateView.as_view(
            crop_url_name=None)(request)

        self.assertEqual(resp.status_code, 200, resp.content)

    def test_cbv_crop_url_conflict(self):
        # target_model not built-in while crop url name is
        # using built-in crop url name
        file = find("demo/screen_upload.png")

        with open(file, "rb") as fp:
            data = {"files[]": fp}
            request = self.factory.post(
                self.get_demo_upload_url(), data=data,
                HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageCreateView.as_view(
                target_model="tests.FakeValidImageModel")(request)

        expected_msg = ("'crop_url_name' in BuiltInImageCreateView "
                        "is using built-in default, while 'target_model' "
                        "is not using built-in default value")
        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_thumbnail_size_invalid(self):
        file = find("demo/screen_upload.png")

        data = {"thumbnail_size": "100.1"}

        with open(file, "rb") as fp:
            data["files[]"] = fp
            request = self.factory.post(
                self.get_demo_upload_url(), data=data,
                HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageCreateView.as_view()(request)

        expected_msg = ("'thumbnail_size' must be an int, or a string of int, "
                        "or a string in the form of '80x60', or a list or tuple of "
                        "2 ints, e.g, [80, 60] or (80, 60).")

        self.assertIn(expected_msg, cm.exception.args[0])


@override_settings(MEDIA_ROOT=test_media_root)
class GalleryWidgetFetchViewTest(ViewTestMixin, TestCase):

    def test_fetch(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        # This was used to make sure only images in the specific
        # gallery will be fetched.
        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        def get_fetched_result():
            resp = self.c.get(self.get_demo_fetch_url(
                params={"pks": list(gallery.images)}),
                HTTP_X_REQUESTED_WITH="XMLHttpRequest")
            self.assertEqual(resp.status_code, 200)
            resp_image_list = [
                file["pk"] for file in json.loads(resp.content)["files"]]
            return resp_image_list

        image_list = list(gallery.images)

        self.c.force_login(self.user)
        self.assertEqual(image_list, get_fetched_result())

        self.c.force_login(self.superuser)
        self.assertEqual(image_list, get_fetched_result())

        another_user = self.create_user()
        self.c.force_login(another_user)
        self.assertEqual(len(get_fetched_result()), 0)

    def test_fetch_thumbnail_size_list(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        self.c.force_login(self.user)

        url = self.get_demo_fetch_url(
            params={"pks": list(gallery.images)})

        url += "&thumbnail_size=100&thumbnail_size=150"

        resp = self.c.get(url, HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 200)

    def test_fetch_without_pks(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        self.c.force_login(self.user)
        resp = self.c.get(self.get_demo_fetch_url(
            params={"not-pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(resp.status_code, 400)

    def test_fetch_fail_not_xml(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        self.c.force_login(self.user)
        resp = self.c.get(self.get_demo_fetch_url(
            params={"not-pks": list(gallery.images)})
        )

        # notice: not 400
        self.assertEqual(resp.status_code, 403)

    def test_fetch_pk_not_digit(self):
        self.c.force_login(self.user)
        resp = self.c.get(self.get_demo_fetch_url(
            params={"pks": [1.1]}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(resp.status_code, 400)

    def test_fetch_pk_decode_error(self):
        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        self.c.force_login(self.user)
        resp = self.c.get(self.get_demo_fetch_url(
            params={"pks": "abcd"}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(resp.status_code, 400)

    def test_fetch_pk_loaded_but_not_list(self):
        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        self.c.force_login(self.user)
        resp = self.c.get(self.get_demo_fetch_url(
            params={"pks": '{"ab": "cd"}'}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(resp.status_code, 400)

    def test_fetch_not_logged_in(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        resp = self.c.get(self.get_demo_fetch_url(
            params={"not-pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(resp.status_code, 302)

    def test_cbv_target_model_not_configured(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        request = self.factory.get(
            self.get_demo_fetch_url(
                params={"pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageListView.as_view(
                target_model=None)(request)

        expected_msg = ("Using BaseImageModelMixin (base class of "
                        "BuiltInImageListView) "
                        "without the 'target_model' attribute is prohibited.")
        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_disable_server_side_crop(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        request = self.factory.get(
            self.get_demo_fetch_url(
                params={"pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        resp = built_in_views.BuiltInImageListView.as_view(
            disable_server_side_crop=True)(request)
        self.assertEqual(resp.status_code, 200)

    def test_cbv_invalid_crop_url(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        request = self.factory.get(
            self.get_demo_fetch_url(
                params={"pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageListView.as_view(
                crop_url_name="invalid-crop-url")(request)

        expected_msg = ("'crop_url_name' in BuiltInImageListView "
                        "is invalid. The exception is: NoReverseMatch: "
                        "Reverse for 'invalid-crop-url' not found.")

        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_crop_url_not_configured(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        request = self.factory.get(
            self.get_demo_fetch_url(
                params={"pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        resp = built_in_views.BuiltInImageListView.as_view(
            crop_url_name=None)(request)

        self.assertEqual(resp.status_code, 200)

    def test_cbv_crop_url_conflict(self):
        # target_model not built-in while crop URL name is
        # using crop url name of built-in image model
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        request = self.factory.get(
            self.get_demo_fetch_url(
                params={"pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageListView.as_view(
                target_model="tests.FakeValidImageModel")(request)

        expected_msg = ("'crop_url_name' in BuiltInImageListView "
                        "is using built-in default, while 'target_model' "
                        "is not using built-in default value")
        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_thumbnail_size_invalid(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        request = self.factory.get(
            self.get_demo_fetch_url(
                params={"pks": list(gallery.images),
                        "thumbnail_size": 100.1}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageListView.as_view()(request)

        expected_msg = ("'thumbnail_size' must be an int, or a string of int, "
                        "or a string in the form of '80x60', or a list of2 ints, "
                        "e.g, [80, 60]. "
                        "Ref: https://stackoverflow.com/a/30107874/3437454")

        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_image_unexpectedly_deleted(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        from sorl.thumbnail import delete

        for obj in BuiltInGalleryImage.objects.all():
            # Why this won't delete the thumbnail cache?
            delete(obj.image.path, delete_file=True)

        request = self.factory.get(
            self.get_demo_fetch_url(
                params={"pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        resp = built_in_views.BuiltInImageListView.as_view()(request)

        self.assertEqual(resp.status_code, 200)

        resp_files = json.loads(resp.content)['files']
        for f in resp_files:
            self.assertEqual(f['error'],
                             'image: The image was unexpectedly deleted from server')

    def test_cbv_get_thumbnail_error(self):
        gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5,
            shuffle=True)

        factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=2,
            shuffle=True)

        request = self.factory.get(
            self.get_demo_fetch_url(
                params={"pks": list(gallery.images)}),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with mock.patch("galleryfield.mixins.get_thumbnail") as mock_get_thumb:
            mock_get_thumb.side_effect = RuntimeError("Unexpected error")
            resp = built_in_views.BuiltInImageListView.as_view()(request)

        self.assertEqual(resp.status_code, 200)

        resp_files = json.loads(resp.content)['files']
        for f in resp_files:
            self.assertEqual(f['error'],
                             'thumbnail: RuntimeError: Unexpected error')


@override_settings(MEDIA_ROOT=test_media_root)
class GalleryWidgetCropViewTest(ViewTestMixin, TestCase):

    def get_crop_post_data(self, **kwargs):
        default_result = {
            "x": 200, "y": 200, "width": 400,
            "height": 810, "rotate": -90}
        default_result.update(**kwargs)
        cropped = self.get_cropped_result(**default_result)
        data = self._get_crop_post_data(cropped_result=cropped)
        return data

    def test_crop_success(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        self.c.force_login(self.user)
        resp = self.c.post(
            self.get_demo_crop_url(pk), data=data,
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 200)

        pk = self.get_demo_crop_pk(1)
        data = self.get_default_crop_post_data()
        self.c.force_login(self.superuser)
        resp = self.c.post(
            self.get_demo_crop_url(pk), data=data,
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 200)

    def test_crop_fail_not_xml(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        self.c.force_login(self.user)
        resp = self.c.post(
            self.get_demo_crop_url(pk), data=data)
        self.assertEqual(resp.status_code, 403)

    def test_crop_png(self):
        self.demo_upload_file(self.user, find("demo/screen_upload.png"))
        factories.DemoGalleryFactory.create(
            creator=self.user, images=[BuiltInGalleryImage.objects.first()])
        data = self.get_crop_post_data()

        self.c.force_login(self.user)
        resp = self.c.post(self.get_demo_crop_url(1), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 200)

    def test_crop_non_exist_image(self):
        data = self.get_crop_post_data()

        self.c.force_login(self.user)
        resp = self.c.post(self.get_demo_crop_url(100), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 404)

    def test_crop_cropResult_invalid(self):
        data = self.get_default_crop_post_data(x=None)

        self.c.force_login(self.user)
        pk = self.get_demo_crop_pk(1)
        resp = self.c.post(self.get_demo_crop_url(pk), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 400)

    def test_crop_no_cropResult(self):
        data = self.get_default_crop_post_data(x=None)

        data.pop("cropped_result")

        self.c.force_login(self.user)
        pk = self.get_demo_crop_pk(1)
        resp = self.c.post(self.get_demo_crop_url(pk), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 400)

    def test_crop_cropResult_not_json(self):
        data = self.get_default_crop_post_data(x=None)

        data["cropped_result"] = {
            "x": 100, "y": 200, "width": 400, "height": 810, "rotate": -90}

        self.c.force_login(self.user)
        pk = self.get_demo_crop_pk(1)
        resp = self.c.post(self.get_demo_crop_url(pk), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 400)

    def test_crop_no_rotate(self):
        data = self.get_default_crop_post_data(rotate=0)

        pk = self.get_demo_crop_pk(1)
        self.c.force_login(self.user)
        resp = self.c.post(self.get_demo_crop_url(pk), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 200, resp.content)

    def test_crop_io_error(self):
        # no actual image file
        gallery = factories.DemoGalleryFactory.create(creator=self.user)
        os.remove(gallery.images.objects.first().image.path)

        pk = gallery.images.objects.first().pk
        data = self.get_crop_post_data()
        self.c.force_login(self.user)
        resp = self.c.post(self.get_demo_crop_url(pk), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 400)

    def test_crop_permission_denied(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        another_user = self.create_user()
        self.c.force_login(another_user)
        resp = self.c.post(self.get_demo_crop_url(pk), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 403)

    def test_crop_not_logged_in(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        resp = self.c.post(self.get_demo_crop_url(pk), data=data,
                           HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(resp.status_code, 302)

    def test_cbv_target_model_not_configured(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        request = self.factory.post(
            self.get_demo_crop_url(pk), data=data,
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        kwargs = {"pk": pk}

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageCropView.as_view(
                target_model=None)(request, **kwargs)

        expected_msg = ("Using BaseImageModelMixin (base class of "
                        "BuiltInImageCropView) "
                        "without the 'target_model' attribute is prohibited.")
        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_disable_server_side_crop(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        request = self.factory.post(
            self.get_demo_crop_url(pk), data=data,
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        kwargs = {"pk": pk}

        with self.assertRaises(SuspiciousOperation) as cm:
            resp = built_in_views.BuiltInImageCropView.as_view(
                disable_server_side_crop=True)(request, **kwargs)
            self.assertEqual(resp.status_code, 400)

        expected_msg = "Server side crop is not enabled"
        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_invalid_crop_url(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        request = self.factory.post(
            self.get_demo_crop_url(pk), data=data,
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        kwargs = {"pk": pk}

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageCropView.as_view(
                crop_url_name="invalid-crop-url")(request, **kwargs)

        expected_msg = ("'crop_url_name' in BuiltInImageCropView "
                        "is invalid. The exception is: NoReverseMatch: "
                        "Reverse for 'invalid-crop-url' not found.")

        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_crop_url_not_configured(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        kwargs = {"pk": pk}
        request = self.factory.post(
            self.get_demo_crop_url(pk), data=data,
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        resp = built_in_views.BuiltInImageCropView.as_view(
            crop_url_name=None)(request, **kwargs)

        self.assertEqual(resp.status_code, 200)

    def test_cbv_crop_url_conflict(self):
        # target_model not built-in while crop url name is
        # using built-in crop url name
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()

        kwargs = {"pk": pk}
        request = self.factory.post(
            self.get_demo_crop_url(pk), data=data,
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageCropView.as_view(
                target_model="tests.FakeValidImageModel")(request, **kwargs)

        expected_msg = ("'crop_url_name' in BuiltInImageCropView "
                        "is using built-in default, while 'target_model' "
                        "is not using built-in default value")
        self.assertIn(expected_msg, cm.exception.args[0])

    def test_cbv_thumbnail_size_invalid(self):
        pk = self.get_demo_crop_pk(0)
        data = self.get_default_crop_post_data()
        data["thumbnail_size"] = 100.1

        kwargs = {"pk": pk}
        request = self.factory.post(
            self.get_demo_crop_url(pk), data=data,
            HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.user = self.user

        with self.assertRaises(ImproperlyConfigured) as cm:
            built_in_views.BuiltInImageCropView.as_view()(request, **kwargs)

        expected_msg = ("'thumbnail_size' must be an int, or a string of int, "
                        "or a string in the form of '80x60', or a list or tuple of "
                        "2 ints, e.g, [80, 60] or (80, 60).")

        self.assertIn(expected_msg, cm.exception.args[0])



import json
from importlib import reload
from unittest import mock

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.forms.renderers import DjangoTemplates
from django.test import SimpleTestCase, TestCase, override_settings
from django.urls import reverse

import galleryfield
from galleryfield import conf, defaults
from galleryfield.fields import GalleryFormField
from galleryfield.utils import get_formatted_thumbnail_size
from galleryfield.widgets import GalleryWidget
from tests import factories
from tests.test_fields import DemoTestGalleryModelForm


class GalleryWidgetTestMixin:
    @staticmethod
    def _get_rendered_field_html(field, print_output=False):
        class Form(forms.Form):
            f = field

        haystack = str(Form()['f'])
        if print_output:
            print(haystack)
        return haystack

    def assertFieldRendersIn(self, field, needle, strict=False, print_output=False):
        haystack = self._get_rendered_field_html(field, print_output)
        assert_in = self.assertInHTML if not strict else self.assertIn
        assert_in(needle, haystack)

    def assertFieldRendersNotIn(self, field, needle, print_output=False):
        haystack = self._get_rendered_field_html(field, print_output)
        self.assertNotIn(needle, haystack)

    def _render_widget(self, widget, name, value=None, attrs=None, **kwargs):
        django_renderer = DjangoTemplates()
        print_output = kwargs.pop("print_output", False)
        output = widget.render(name, value, attrs=attrs,
                               renderer=django_renderer, **kwargs)
        if print_output:
            print(output)
        return output

    def check_in_html(self, widget, name, value, html, attrs=None,
                      strict=False, print_output=False, **kwargs):
        output = self._render_widget(widget, name, value, attrs=attrs, **kwargs)
        assert_in = self.assertIn if strict else self.assertInHTML

        if print_output:
            print(output)

        if isinstance(html, str):
            html = [html]
        for _html in html:
            assert_in(_html, output)

    def check_not_in_html(self, widget, name, value, html, attrs=None, **kwargs):
        output = self._render_widget(widget, name, value, attrs=attrs, **kwargs)
        if isinstance(html, str):
            html = [html]
        for _html in html:
            self.assertNotIn(_html, output)


class GalleryWidgetTest(GalleryWidgetTestMixin, SimpleTestCase):

    def test_widget(self):
        field = GalleryFormField()
        self.assertIsInstance(field.widget, GalleryWidget)

    def test_required_widget_render(self):
        f = GalleryFormField(required=True)

        self.assertFieldRendersIn(
            f, '<input type="hidden" name="f" value="null"'
               ' class="django-galleryfield-files-field '
               ' hiddeninput" required id="id_f">')

        f = GalleryFormField(required=False)
        self.assertFieldRendersIn(
            f, '<input type="hidden" name="f" value="null"'
               ' class="django-galleryfield-files-field '
               ' hiddeninput" id="id_f">')

    def test_gallery_widget_render_with_value(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)
        image_data = [1]
        value = json.dumps(image_data)
        expected_result = (
            '<input type="hidden" name="image" value="[1]"')
        self.check_in_html(
            f.widget, "image", value, strict=True,  html=[expected_result])

    def test_gallery_widget_jquery_upload_options_max_number_of_files_overridden(self):  # noqa
        from random import randint
        max_number_of_file_ui_options_value = randint(1, 10)
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        f.widget.jquery_file_upload_ui_options.update(
            {"maxNumberOfFiles": max_number_of_file_ui_options_value})

        setattr(f.widget, "max_number_of_images", None)
        self.check_not_in_html(f.widget, "image", '', html="maxNumberOfFiles")

        f.widget = GalleryWidget(
            jquery_file_upload_ui_options={
                "maxNumberOfFiles": max_number_of_file_ui_options_value})
        setattr(f.widget, "max_number_of_images", 0)
        self.check_not_in_html(f.widget, "image", '', html="maxNumberOfFiles")

        max_number_of_file = randint(1, 10)
        f.widget = GalleryWidget(
            jquery_file_upload_ui_options={"maxNumberOfFiles": 0})

        setattr(f.widget, "max_number_of_images", max_number_of_file)
        expected_string = "maxNumberOfFiles: %i" % max_number_of_file
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)

    def test_gallery_widget_thumbnail_size(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)
        f.widget = GalleryWidget()
        expected_value = get_formatted_thumbnail_size(
                conf.DEFAULT_THUMBNAIL_SIZE).split("x")[0]
        expected_string = f"previewMaxWidth: {expected_value}"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)

        f.widget = GalleryWidget(thumbnail_size=130)
        expected_string = "previewMaxWidth: 130"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)
        expected_string = "previewMaxHeight: 130"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)

        f.widget = GalleryWidget(thumbnail_size=(135, 250))
        expected_string = "previewMaxWidth: 135"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)
        expected_string = "previewMaxHeight: 250"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)

        f.widget = GalleryWidget(thumbnail_size="130x260")
        expected_string = "previewMaxWidth: 130"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)
        expected_string = "previewMaxHeight: 260"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)

    def test_set_thumbnail_size_after_gallery_widget_init(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        f.widget = GalleryWidget(thumbnail_size=130)
        f.widget.thumbnail_size = 250
        expected_string = "previewMaxWidth: 250"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)
        expected_string = "previewMaxHeight: 250"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)

        f.widget.thumbnail_size = "111x222"
        expected_string = "previewMaxWidth: 111"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)
        expected_string = "previewMaxHeight: 222"
        self.check_in_html(f.widget, "image", '', strict=True, html=expected_string)

    def test_gallery_widget_jquery_upload_options_None(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)
        self.check_in_html(
            f.widget, "image", '', strict=True, html="disableImageResize")

        f.widget = GalleryWidget(
            jquery_file_upload_ui_options={"disableImageResize": None})
        self.check_not_in_html(f.widget, "image", '', html="disableImageResize")

    def test_gallery_widget_disabled(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)
        f.widget = GalleryWidget()
        file_upload_button = (
            '<input type="file" class="gallery-widget-image-input" '
            'id="%(field_name)s-files" multiple accept="image/*" '
            'data-action="%(upload_url)s">'
            % {"field_name": "image",
               "upload_url":
                   reverse(defaults.DEFAULT_UPLOAD_URL_NAME)}
        )
        self.check_in_html(
            f.widget, "image", '',
            html=[file_upload_button])

        f.widget.attrs["readonly"] = True
        self.check_not_in_html(
            f.widget, "image", '',
            # The css class of file input button
            html=["gallery-widget-image-input"])

    def test_gallery_widget_upload_handler_url_none(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)
        f.widget = GalleryWidget()
        file_upload_button = (
            '<input type="file" class="gallery-widget-image-input" '
            'id="%(field_name)s-files" multiple accept="image/*" '
            'data-action="%(upload_url)s">'
            % {"field_name": "image",
               "upload_url":
                   reverse(defaults.DEFAULT_UPLOAD_URL_NAME)}
        )
        self.check_in_html(
            f.widget, "image", '',
            html=[file_upload_button])

        f.widget.upload_url = None
        self.check_not_in_html(
            f.widget, "image", '',
            # The css class of file input button
            html=["gallery-widget-image-input"])

    def test_disabled_widget_render(self):
        f = GalleryFormField()
        self.assertFieldRendersIn(
            f, 'gallery-widget-image-input', strict=True, print_output=True)

        f = GalleryFormField(disabled=True)
        self.assertFieldRendersNotIn(f, 'gallery-widget-image-input')

    def test_widget_render_conflict(self):
        # the target image model is not the default,
        # some of the urls are default urls
        field = GalleryFormField(target_model="tests.FakeValidImageModel")

        test_case = {
            "galleryfield-builtingalleryimage-upload":
                {"upload_url": "test_image_upload"},
            "galleryfield-builtingalleryimage-crop":
                {"crop_request_url": "test_image_crop"},
            "galleryfield-builtingalleryimage-fetch":
                {"fetch_url": "test_images_fetch"}
        }

        default_urls = {
            "upload_url": "galleryfield-builtingalleryimage-upload",
            "crop_request_url": "galleryfield-builtingalleryimage-crop",
            "fetch_url": "galleryfield-builtingalleryimage-fetch"
        }

        for default, kwargs in test_case.items():
            with self.subTest(default=default, url_kwargs=kwargs):
                test_kwargs = default_urls.copy()
                test_kwargs.update(kwargs)

                field.widget = GalleryWidget(**test_kwargs)
                with self.assertRaises(ImproperlyConfigured) as cm:
                    self._render_widget(field.widget, "field", "")

                self.assertIn(
                    'You need to write your own views for your image model',
                    cm.exception.args[0], cm.exception)
                self.assertNotIn(
                    default,
                    cm.exception.args[0]
                )

    def test_widget_disable_fetch_no_conflict(self):
        field = GalleryFormField(target_model="tests.FakeValidImageModel")

        kwargs = {
            "upload_url": "test_image_upload",
            "fetch_url": "test_images_fetch",
            "disable_server_side_crop": True
        }

        field.widget = GalleryWidget(**kwargs)
        self._render_widget(field.widget, "field", "")

    def test_widget_disable_server_side_crop_no_conflict(self):
        field = GalleryFormField(target_model="tests.FakeValidImageModel")

        kwargs = {
            "upload_url": "test_image_upload",
            "fetch_url": "builtingalleryimage-fetch",  # a conflict URL
            "disable_fetch": True
        }

        field.widget = GalleryWidget(**kwargs)
        self._render_widget(field.widget, "field", "")

    def test_widget_no_conflict(self):
        # the target image model and all urls are not using the default,
        field = GalleryFormField(target_model="tests.FakeValidImageModel")

        kwargs = {
            "upload_url": "test_image_upload",
            "crop_request_url": "test_image_crop",
            "fetch_url": "test_images_fetch"
        }

        field.widget = GalleryWidget(**kwargs)
        # No error thrown
        self._render_widget(field.widget, "field", "")

    def test_widget_invalid_url(self):
        # the target image model and all urls are not using the default,
        field = GalleryFormField(target_model="tests.FakeValidImageModel")

        kwargs = {
            "upload_url": "test_image_upload",
            "fetch_url": "test_images_fetch"
        }

        invalid_url_name = "invalid-url-name"

        for k, v in kwargs.items():
            with self.subTest(key=k, value=v):
                test_kwargs = kwargs.copy()
                test_kwargs.update({k: invalid_url_name})

                field.widget = GalleryWidget(**test_kwargs)
                with self.assertRaises(ImproperlyConfigured) as cm:
                    self._render_widget(field.widget, "field", "")

                expected_error_str = "is not a valid view function or pattern name"

                self.assertIn(
                    expected_error_str,
                    str(cm.exception))

    def test_widget_set_jquery_file_upload_ui_options_None_get_default(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        f.widget.jquery_file_upload_ui_options = None
        self.assertDictEqual(
            f.widget.jquery_file_upload_ui_options,
            defaults.JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS)

    def test_widget_set_jquery_file_upload_ui_options_not_dict(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        with self.assertRaises(ImproperlyConfigured) as cm:
            f.widget.jquery_file_upload_ui_options = "foo-bar"

        expected_error_msg = "'jquery_file_upload_ui_options' must be a dict"
        self.assertIn(expected_error_msg, cm.exception.args[0])

    @mock.patch("django.forms.renderers.DjangoTemplates.render")
    def test_widget_set_jquery_file_upload_ui_options_render(self, mock_render):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        f.widget.jquery_file_upload_ui_options = {
            "autoUpload": True
        }
        self._render_widget(f.widget, "image")

        expected_string = "autoUpload: true"
        self.assertIn(
            expected_string,
            str(mock_render.call_args),
        )

    @mock.patch('galleryfield.widgets.logger.warning')
    @mock.patch("django.forms.renderers.DjangoTemplates.render")
    def test_widget_set_jquery_file_upload_ui_options_configured_maxNumberOfFiles(
            self, mock_render, mock_log):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        f.widget.jquery_file_upload_ui_options = {
            "maxNumberOfFiles": 10
        }

        expected_message = ("'maxNumberOfFiles' in 'jquery_file_upload_ui_options' "
                            "will be overridden later by the formfield")
        self.assertEqual(mock_log.call_count, 1)
        self.assertIn(expected_message, str(mock_log.call_args))
        self._render_widget(f.widget, "image")
        self.assertNotIn(
            "maxNumberOfFiles",
            str(mock_render.call_args.kwargs),
        )

    @mock.patch('galleryfield.widgets.logger.warning')
    @mock.patch("django.forms.renderers.DjangoTemplates.render")
    def test_widget_set_jquery_file_upload_ui_options_configured_preview_size(  # noqa
            self, mock_render, mock_log):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        values = [False, "false", "False"]

        for value in values:
            with self.subTest(value=value):

                f.widget.jquery_file_upload_ui_options = {
                    "singleFileUploads": value
                }

                expected_message = ("'singleFileUploads=False' in "
                                    "'jquery_file_upload_ui_options' is not "
                                    "allowed and will be ignored.")

                # self.assertEqual(mock_log.call_count, 1, list(mock_log.call_args))
                self.assertIn(expected_message, str(mock_log.call_args))
                self._render_widget(f.widget, "image")
                self.assertNotIn(
                    "singleFileUploads",
                    str(mock_render.call_args),
                )
                mock_render.reset_mock()
                mock_log.reset_mock()

    @mock.patch('galleryfield.widgets.logger.warning')
    @mock.patch("django.forms.renderers.DjangoTemplates.render")
    def test_widget_set_jquery_file_upload_ui_options_configured_singleFileUploads_true(  # noqa
            self, mock_render, mock_log):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)
        f.widget.thumbnail_size = "120x180"

        values = [{"previewMaxWidth": 100},
                  {"previewMaxHeight": 150}]

        expected_rendered_values = [
            "previewMaxWidth: 120",
            "previewMaxHeight: 180"]

        for i, value in enumerate(values):
            with self.subTest(value=value):

                f.widget.jquery_file_upload_ui_options = value

                expected_message = (
                    "'previewMaxWidth' and 'previewMaxHeight' "
                    "in 'jquery_file_upload_ui_options' are ignored.")
                # self.assertEqual(mock_log.call_count, 1)
                self.assertIn(expected_message, str(mock_log.call_args))
                self._render_widget(f.widget, "image")
                self.assertIn(
                    expected_rendered_values[i],
                    str(mock_render.call_args),
                )
                mock_render.reset_mock()
                mock_log.reset_mock()

    def test_widget_set_jquery_file_upload_ui_sortable_options_None_get_default(self):  # noqa
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        f.widget.jquery_file_upload_ui_sortable_options = None
        self.assertDictEqual(
            f.widget.jquery_file_upload_ui_sortable_options,
            defaults.JQUERY_FILE_UPLOAD_UI_DEFAULT_SORTABLE_OPTIONS)

    def test_widget_set_jquery_file_upload_ui_sortable_options_not_dict(self):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        with self.assertRaises(ImproperlyConfigured) as cm:
            f.widget.jquery_file_upload_ui_sortable_options = "foo-bar"

        expected_error_msg = (
            "'jquery_file_upload_ui_sortable_options' must be a dict")
        self.assertIn(expected_error_msg, cm.exception.args[0])

    @mock.patch("django.forms.renderers.DjangoTemplates.render")
    def test_widget_set_jquery_file_upload_ui_sortable_options_render(
            self, mock_render):
        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)

        f.widget.jquery_file_upload_ui_sortable_options = {
            "delay": 400
        }
        self._render_widget(f.widget, "image")

        expected_string = "delay: 400"
        self.assertIn(
            expected_string,
            str(mock_render.call_args),
        )

        self.assertIn(
            "disabled: false",
            str(mock_render.call_args),
        )


class GalleryWidgetConfTest(GalleryWidgetTestMixin, SimpleTestCase):
    @staticmethod
    def reload_defaults_and_conf():
        reload(galleryfield.defaults)
        reload(galleryfield.conf)

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG={"bootstrap_version": 5})
    @mock.patch("django.forms.renderers.DjangoTemplates.render")
    def test_widget_bootstrap_version_render(self, mock_render):
        self.reload_defaults_and_conf()

        f = GalleryFormField(target_model=defaults.DEFAULT_TARGET_IMAGE_MODEL)
        self._render_widget(f.widget, "image")
        expected_string = "showElementClass: 'show'"
        self.assertIn(
            expected_string,
            str(mock_render.call_args),
        )

    def test_conf_static_with_different_bootstrap_versions(self):
        for bootstrap_version in [3, 4, 5]:
            with override_settings(
                    DJANGO_GALLERY_FIELD_CONFIG={
                        "bootstrap_version": bootstrap_version}):
                self.reload_defaults_and_conf()

                self.assertEqual(
                    conf.BOOTSTRAP_CSS_LOCATION,
                    defaults.DEFAULT_STATICS["bootstrap_css"][bootstrap_version])

                self.assertEqual(
                    conf.BOOTSTRAP_JS_LOCATION,
                    defaults.DEFAULT_STATICS["bootstrap_js"][bootstrap_version])


class GalleryWidgetTestExtra(TestCase):
    # This test cases need db support
    def setUp(self) -> None:
        factories.UserFactory.reset_sequence()
        factories.BuiltInGalleryImageFactory.reset_sequence()
        factories.DemoGalleryFactory.reset_sequence()
        self.user = factories.UserFactory()
        super().setUp()

    def test_no_fetch_url(self):
        gallery_obj = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5, shuffle=True)
        pks = list(gallery_obj.images)

        form = DemoTestGalleryModelForm(instance=gallery_obj)

        rendered_js_content = "// fetching existing images"
        rendered_js_instance_data = f"({pks})"
        self.assertIn(rendered_js_content, form.as_table())
        self.assertIn(rendered_js_instance_data, form.as_table())

        # now we set fetch_url=None to the widget
        form.fields["images"].widget.fetch_url = None

        self.assertNotIn(rendered_js_content, form.as_table())
        self.assertNotIn(rendered_js_instance_data, form.as_table())

    def test_no_crop_disabled(self):
        gallery_obj = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5, shuffle=True)

        form = DemoTestGalleryModelForm(instance=gallery_obj)

        # This data attribute only exists in Edit buttons
        rendered_button_data_toggle = 'data-toggle="modal"'
        self.assertIn(rendered_button_data_toggle, form.as_table())

        # # now we set crop_request_url=None to the widget
        form.fields["images"].widget.disable_server_side_crop = True
        self.assertNotIn(rendered_button_data_toggle, form.as_table())



JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS = {
    "autoUpload": False,
    "imageMaxWidth": 1024,
    "imageMaxHeight": 1024,
    "loadImageFileTypes": r"/^image\/(gif|jpeg|png|bmp|webp|svg\+xml|x-icon)$/",
    "sequentialUploads": True,
    "acceptFileTypes": r"/(\.|\/)(png|gif|bmp|jpe?g|tif|ico|webp)$/i",
    "imageOrientation": True,
    "maxFileSize": 1.5 * 1024 ** 2,  # 1.5Mb
    "minFileSize": 0.0001 * 1024 ** 2,  # 0.0001Mb
    "disableImageResize": "/Android(?!.*Chrome)|Opera/.test(window.navigator "
                          "&& navigator.userAgent)",
}
"""dict: The default options for jQuery-File-Upload module in jQuery file upload UI.
The values can be overridden when initializing
:class:`galleryfield.widgets.GalleryWidget`
via :attr:`jquery_file_upload_ui_options`.
Please refer to `available options
<https://github.com/blueimp/jQuery-File-Upload/wiki/Options#general-options>`__
for the details and more options.
"""


JQUERY_FILE_UPLOAD_UI_DEFAULT_SORTABLE_OPTIONS = {
    "disabled": False,
    "delay": 300,
    "animation": 200,
    "touchStartThreshold": 5,
    "ghostClass": "galleryWidget-sortable-ghost",
    "chosenClass": "galleryWidget-sortable-chosen",
    "filter": ".btn, .toggle, img, a, span, progress",
}
"""dict: The default options for SortableJS used in jQuery file upload UI. See
`available options <https://github.com/SortableJS/Sortable#options>`__
for details. Notice that, callback options currently can not be configured
here.
"""


# {{{ DO NOT change this if you have deployed this app on you production server!
# Or else it might cause unexpected result!

DEFAULT_UPLOAD_URL_NAME = "galleryfield-builtingalleryimage-upload"
DEFAULT_CROP_URL_NAME = "galleryfield-builtingalleryimage-crop"
DEFAULT_FETCH_URL_NAME = "galleryfield-builtingalleryimage-fetch"

DEFAULT_TARGET_IMAGE_MODEL = "galleryfield.BuiltInGalleryImage"
DEFAULT_TARGET_IMAGE_FIELD_NAME = "image"
DEFAULT_CREATOR_FIELD_NAME = "creator"

# }}}

BUILT_IN_JS = ["js/galleryfield-ui.js"]
BUILT_IN_CSS = []

EXTRA_JS = []
EXTRA_CSS = []

DEFAULT_THUMBNAIL_SIZE = "120x120"
DEFAULT_THUMBNAIL_QUALITY = 80

WIDGET_HIDDEN_INPUT_CSS_CLASS = "django-galleryfield"

PROMPT_ALERT_ON_WINDOW_RELOAD_IF_CHANGED = True

DEFAULT_BOOTSTRAP_VERSION = 3

_BOOTSTRAP_CSS_PATTERN = (
    "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/"
    "{}/css/bootstrap.min.css")

_BOOTSTRAP_JS_PATTERN = (
    "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/"
    "{}/js/bootstrap.min.js")

_JQUERY_LOCATION = (
    "https://cdnjs.cloudflare.com/ajax/libs/jquery/3.6.0/jquery.min.js")

DEFAULT_STATICS = {
    "jquery": _JQUERY_LOCATION,
    "bootstrap_css": {3: _BOOTSTRAP_CSS_PATTERN.format("3.4.1"),
                      4: _BOOTSTRAP_CSS_PATTERN.format("4.6.1"),
                      5: _BOOTSTRAP_CSS_PATTERN.format("5.1.3"),
                      },
    "bootstrap_js": {3: _BOOTSTRAP_JS_PATTERN.format("3.4.1"),
                     4: _BOOTSTRAP_JS_PATTERN.format("4.6.1"),
                     5: _BOOTSTRAP_JS_PATTERN.format("5.1.3")},
}
"""dict: The default assets used to render the :class:`GalleryWidget` instance.
"""



import random
from unittest import mock

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.forms import ValidationError
from django.forms.widgets import Textarea
from django.test import TestCase
from django.test.utils import isolate_apps, override_settings

from demo.models import DemoGallery
from galleryfield.fields import GalleryField, GalleryFormField
from galleryfield.widgets import GalleryWidget
from tests import factories
from tests.factories import DemoGalleryFactory


class DemoTestGalleryModelForm(forms.ModelForm):
    class Meta:
        model = DemoGallery
        fields = ["images"]


class GalleryFieldTest(TestCase):
    def setUp(self) -> None:
        factories.UserFactory.reset_sequence()
        factories.BuiltInGalleryImageFactory.reset_sequence()
        factories.DemoGalleryFactory.reset_sequence()
        self.user = factories.UserFactory()
        super().setUp()

    def test_form_save(self):
        image = factories.BuiltInGalleryImageFactory(creator=self.user)
        form = DemoTestGalleryModelForm(
            data={"images": [image.pk]})
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(DemoGallery.objects.count(), 1)

    def test_form_add_images(self):
        instance = factories.DemoGalleryFactory.create(creator=self.user)
        image2 = factories.BuiltInGalleryImageFactory(creator=self.user)
        self.assertEqual(DemoGallery.objects.count(), 1)
        self.assertEqual(len(DemoGallery.objects.first().images), 1)

        form = DemoTestGalleryModelForm(
            data={"images": instance.images + [image2.pk]},
            instance=instance
        )
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(DemoGallery.objects.count(), 1)
        new_images = DemoGallery.objects.first().images
        self.assertEqual(len(new_images), 2)

    def test_form_change_images(self):
        instance = DemoGalleryFactory.create(creator=self.user)
        self.assertEqual(DemoGallery.objects.count(), 1)
        self.assertEqual(len(DemoGallery.objects.first().images), 1)
        image2 = factories.BuiltInGalleryImageFactory(creator=self.user)

        form = DemoTestGalleryModelForm(
            data={"images": [image2.pk]},
            instance=instance
        )
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(DemoGallery.objects.count(), 1)
        new_images = DemoGallery.objects.first().images
        self.assertEqual(len(new_images), 1)

    def test_form_save_null(self):
        form = DemoTestGalleryModelForm(
            data={"images": ''},
        )
        form.fields["images"].required = False

        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(DemoGallery.objects.count(), 1)
        new_images = DemoGallery.objects.first().images
        self.assertEqual(new_images, [])

    def test_form_replace_null(self):
        instance = DemoGalleryFactory.create(creator=self.user)
        self.assertEqual(DemoGallery.objects.count(), 1)
        self.assertEqual(len(DemoGallery.objects.first().images), 1)

        form = DemoTestGalleryModelForm(
            data={"images": ''},
            instance=instance
        )
        form.fields["images"].required = False

        self.assertTrue(form.is_valid(), form.errors)
        form.save()

        self.assertEqual(DemoGallery.objects.count(), 1)
        new_images = DemoGallery.objects.first().images
        self.assertEqual(new_images, [])

    def test_form_invalid(self):
        instance = DemoGalleryFactory.create(creator=self.user)
        self.assertEqual(DemoGallery.objects.count(), 1)
        self.assertEqual(len(DemoGallery.objects.first().images), 1)

        form = DemoTestGalleryModelForm(
            data={"images": ''},
            instance=instance
        )
        form.fields["images"].required = True

        self.assertFalse(form.is_valid())


class GalleryFormFieldTest(TestCase):
    def setUp(self) -> None:
        factories.UserFactory.reset_sequence()
        factories.BuiltInGalleryImageFactory.reset_sequence()
        factories.DemoGalleryFactory.reset_sequence()
        self.user = factories.UserFactory()
        super().setUp()

    def test_gallery_form_field_clean_image_not_exist(self):
        field = GalleryFormField()
        form_data = [100]

        msg = "'The submitted file is empty.'"

        with self.assertRaisesMessage(ValidationError, msg):
            field.clean(form_data)

    def test_gallery_form_field_clean(self):
        image = factories.BuiltInGalleryImageFactory(creator=self.user)
        field = GalleryFormField()
        form_data = [image.pk]
        cleaned_data = field.clean(form_data)
        self.assertEqual(cleaned_data, form_data)

    def test_gallery_form_field_clean_null_required(self):
        field = GalleryFormField(required=True)
        inputs = [
            '',
            [],
        ]

        msg = "'The submitted file is empty.'"

        for data in inputs:
            with self.subTest(data=data):
                with self.assertRaisesMessage(ValidationError, msg):
                    field.clean(data)

    def test_gallery_form_field_clean_null_not_required(self):
        field = GalleryFormField(required=False)
        inputs = [
            '',
            None,
            'null',
            []
        ]

        for data in inputs:
            with self.subTest(data=data):
                self.assertIsNone(field.clean(data))

    def test_gallery_form_field_clean_invalid_image_json(self):
        inputs = ['invalid-image']
        msg = "The submitted images are invalid."

        for required in [True, False]:
            with self.subTest(required=required):
                field = GalleryFormField(required=required)
                with self.assertRaisesMessage(ValidationError, msg):
                    field.clean(inputs)

    def test_gallery_form_field_clean_not_null_not_list(self):
        input_str = 'invalid-image'
        msg = "The submitted images are invalid."

        for required in [True, False]:
            with self.subTest(required=required):
                field = GalleryFormField(required=required)
                with self.assertRaisesMessage(ValidationError, msg):
                    field.clean(input_str)

    def test_gallery_form_field_clean_disabled_invalid(self):
        field = GalleryFormField(disabled=True)
        input_str = 'invalid-image'
        msg = "The submitted images are invalid."

        with self.assertRaisesMessage(ValidationError, msg):
            field.clean(input_str)

    def test_gallery_form_field_assign_max_number_of_images(self):
        field = GalleryFormField(required=False)
        max_number_of_images_list = [
            0,
            "1",
            "123",
            1234,
            None
        ]

        for n in max_number_of_images_list:
            with self.subTest(max_number_of_images=n):
                field.max_number_of_images = n
                if n is not None:
                    self.assertEqual(field.max_number_of_images, int(n))
                    self.assertEqual(field.widget.max_number_of_images, int(n))
                else:
                    self.assertIsNone(field.max_number_of_images)
                    self.assertIsNone(field.widget.max_number_of_images)

    def test_gallery_form_field_assign_max_number_of_images_invalid(self):
        field = GalleryFormField(required=False)
        max_number_of_images_list = [
            -1,
            "-1",
            "abc",
            object,
        ]

        for n in max_number_of_images_list:
            with self.subTest(max_number_of_images=n):
                with self.assertRaises(TypeError):
                    field.max_number_of_images = n

    def test_gallery_form_field_clean_max_number_of_images_exceeded(self):
        field = GalleryFormField()
        n = 1
        field.max_number_of_images = n

        images = factories.BuiltInGalleryImageFactory.create_batch(
            size=2, creator=self.user)

        msg = "Number of images exceeded, only %i allowed" % n

        with self.assertRaisesMessage(ValidationError, msg):
            field.clean([images[0].pk, images[1].pk])

    def test_gallery_form_field_clean_max_number_of_images_not_exceeded(self):
        field = GalleryFormField()
        field.max_number_of_images = 1

        image = factories.BuiltInGalleryImageFactory(creator=self.user)

        self.assertEqual(field.clean([image.pk]), [image.pk])

    def test_gallery_form_field_clean_max_number_of_images_zero(self):
        # zero means not limited
        field = GalleryFormField()
        field.max_number_of_images = 0

        images = factories.BuiltInGalleryImageFactory.create_batch(
            size=2, creator=self.user)

        data = [images[0].pk, images[1].pk]

        cleaned_data = field.clean(data)
        self.assertEqual(cleaned_data, data)

    def test_gallery_form_field_target_image_model_valid(self):
        GalleryFormField(target_model="tests.FakeValidImageModel")

    def test_gallery_form_field_target_image_model_invalid(self):
        with self.assertRaises(ImproperlyConfigured):
            GalleryFormField(target_model="tests.FakeInvalidImageModel1")

    @mock.patch('galleryfield.fields.logger.info')
    def test_gallery_form_field_target_image_model_configured_none(self, mock_log):
        GalleryFormField()
        self.assertEqual(mock_log.call_count, 1)

    @mock.patch('galleryfield.fields.logger.info')
    @override_settings(SILENCED_SYSTEM_CHECKS=["gallery_form_field.I001"])
    def test_gallery_form_field_target_image_model_configured_none_log_suppressed(self, mock_log):  # noqa
        GalleryFormField()
        self.assertEqual(mock_log.call_count, 0)

    def test_gallery_form_field_changed_widget(self):
        max_number_of_images = 5
        field = GalleryFormField(
            target_model="tests.FakeValidImageModel",
            max_number_of_images=max_number_of_images,
            required=False)
        old_widget = field.widget
        self.assertEqual(old_widget.max_number_of_images, max_number_of_images)

        # Before assigning to the formfield
        new_widget = GalleryWidget()
        self.assertNotEqual(new_widget.attrs, old_widget.attrs)
        self.assertFalse(hasattr(new_widget, "image_model"))
        self.assertFalse(hasattr(new_widget, "max_number_of_images"))

        # New widget will get updated
        field.widget = new_widget
        self.assertEqual(field.widget.max_number_of_images, max_number_of_images)
        self.assertEqual(field.widget.image_model, old_widget.image_model)
        self.assertEqual(field.widget.widget_is_servicing, field.__class__.__name__)
        self.assertEqual(field.widget.attrs, old_widget.attrs)
        self.assertEqual(field.widget.is_required, False)

    def test_gallery_form_field_textarea_widget_attribute_updated(self):
        form = DemoTestGalleryModelForm()

        max_n_of_images = random.randint(6, 10)

        # We set the max_number_of_images before change the widget,
        # to test after widget change, the attribute also exists
        form.fields["images"].max_number_of_images = max_n_of_images

        form.fields["images"].widget = Textarea()

        self.assertEqual(
            form.fields["images"].widget.max_number_of_images, max_n_of_images)

    def test_gallery_form_field_textarea_widget_max_number_of_images_validate(self):
        # create a gallery with 5 images pk randomized
        my_gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5, shuffle=True)

        form = DemoTestGalleryModelForm(data={"images": my_gallery.images})

        max_n_of_images = 2
        form.fields["images"].max_number_of_images = max_n_of_images
        form.fields["images"].widget = Textarea()

        self.assertFalse(form.is_valid())
        self.assertIn("Number of images exceeded, only 2 allowed", str(form.errors))

    def test_gallery_form_field_textarea_widget_value_render(self):
        # create a gallery with 5 images pk randomized
        my_gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5, shuffle=True)

        form = DemoTestGalleryModelForm(instance=my_gallery)

        image_pks = my_gallery.images
        self.assertIn(str(image_pks), form.as_table())

    def test_gallery_form_field_textarea_widget_value_sequence(self):
        # create a gallery with 5 images pk randomized
        my_gallery = factories.DemoGalleryFactory.create(
            creator=self.user, number_of_images=5, shuffle=True)

        image_pks = my_gallery.images
        original_image_pks = image_pks[:]

        # Here we shuffle the image_pk, and submit it as the form data
        while True:
            random.shuffle(image_pks)
            if image_pks != original_image_pks:
                break

        form = DemoTestGalleryModelForm(data={"images": image_pks})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["images"], image_pks)


class GalleryFieldCheckTest(TestCase):
    def test_field_checks_valid(self):

        class MyModel(models.Model):
            field = GalleryField(target_model="tests.FakeInvalidImageModel1")

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "gallery_field.E004")

    @isolate_apps("tests")
    def test_field_checks_use_default_target(self):

        class MyModel(models.Model):
            field = GalleryField()

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "gallery_field.I001")

    @isolate_apps("tests")
    def test_field_checks_use_invalid_target(self):

        class MyModel(models.Model):
            field = GalleryField(target_model="non-exist.model")

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "gallery_field.E002")

    @isolate_apps("tests")
    def test_field_checks_use_invalid_target_object(self):
        from tests.models import FakeValidImageModel

        class MyModel(models.Model):
            field = GalleryField(target_model=FakeValidImageModel)

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "gallery_field.E001")

    @isolate_apps("tests")
    def test_field_checks_use_invalid_get_image_field_method(self):

        class MyModel(models.Model):
            field = GalleryField(target_model="tests.FakeInvalidImageModel5")

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "gallery_field.E003")

    @isolate_apps("tests")
    def test_field_checks_use_get_image_field_method_not_callable(self):

        class MyModel(models.Model):
            field = GalleryField(target_model="tests.FakeInvalidImageModel4")

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "gallery_field.E005")

    @isolate_apps("tests")
    def test_field_checks_use_get_image_field_method_not_imagefield(self):

        class MyModel(models.Model):
            field = GalleryField(target_model="tests.FakeInvalidImageModel3")

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "gallery_field.E005")

    @isolate_apps("tests")
    # Here we are not loading "demo", while use MyImageModel
    @override_settings(INSTALLED_APPS=['galleryfield', 'tests'])
    def test_field_checks_app_not_in_installed_apps(self):
        class MyModel(models.Model):
            field = GalleryField("demo.MyImageModel")

        # We want to make sure it won't raise error when initializing.
        model = MyModel()

        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "gallery_field.E002")
        self.assertIn("LookupError", errors[0].msg)



from django.core.exceptions import PermissionDenied
from django.views.generic import CreateView, UpdateView
from django.views.generic.list import BaseListView

from galleryfield.mixins import (BaseCreateMixin, BaseCropViewMixin,
                                 BaseListViewMixin)


class ImageCreateView(BaseCreateMixin, CreateView):
    """
    The Class-based view handling the saving of uploaded image.

    .. attribute:: target_model

       |view_target_model|

    .. attribute:: crop_url_name

       |view_crop_url_name|

    .. attribute:: disable_server_side_crop

       |view_disable_server_side_crop|

    .. automethod:: create_instance_from_form
    """
    def form_valid(self, form):
        self.create_instance_from_form(form)
        return super().form_valid(form)

    def create_instance_from_form(self, form):
        """User should provide this method to save the object which will be used 
        in form_valid method. Typically, ``self.object`` is expected to be saved 
        here with  
        `the-save-method <https://docs.djangoproject.com/en/dev/topics/forms/modelforms/#the-save-method>`_.

        See :class:`galleryfield.image_views.BuiltInImageCreateView` for example.
        """  # noqa

        raise NotImplementedError(
            'subclasses of ImageCreateView must provide '
            'a create_instance_from_form() method')


class ImageListView(BaseListViewMixin, BaseListView):
    """
    The Class-based view for fetching the existing images of the gallery instance.

    .. attribute:: target_model

       |view_target_model|

    .. attribute:: crop_url_name

       |view_crop_url_name|

    .. attribute:: disable_server_side_crop

       |view_disable_server_side_crop|

    .. automethod:: get_queryset
    """
    def get_queryset(self):
        """
        User need to override this method to do some basic filter in terms of
        who can see which images.

        :return: A Queryset
        """
        return super().get_queryset()


class ImageCropView(BaseCropViewMixin, UpdateView):
    """
    The Class-based view handling server side cropping of an image model
    instance. Note that a new image model instance will be created rather
    than updating the cropped instance.

    .. attribute:: target_model

       |view_target_model|

    .. attribute:: crop_url_name

       |view_crop_url_name|

    .. attribute:: disable_server_side_crop

       |view_disable_server_side_crop|

    .. automethod:: create_cropped_instance_from_form
    """
    def form_valid(self, form):
        """Save the object from the form. This method should only be overridden
        when the model contains dynamic fields which need to be updated. For
        example, when there is a  ``DateTimeField`` which records the updated
        datetime of the image.
        """
        self.create_cropped_instance_from_form(form)
        return super().form_valid(form)

    def create_cropped_instance_from_form(self, form):
        """User should provide this method to save the object which will be used 
        in form_valid method. Typically, self.object is expected to be saved 
        here with  
        `the-save-method <https://docs.djangoproject.com/en/dev/topics/forms/modelforms/#the-save-method>`_.

        See :class:`galleryfield.image_views.BuiltInImageCropView` for example.
        
        Note that beside the id (pk) and image field, all other field values of the
        saved cropped instance will be the same with the original instance. You 
        need to set those values before calling ``self.object.save()`` method.
        
        """  # noqa

        raise NotImplementedError(
            'subclasses of ImageCropView must provide '
            'a create_cropped_instance_from_form() method')


class BuiltInImageCreateView(ImageCreateView):
    target_model = "galleryfield.BuiltInGalleryImage"
    crop_url_name = "galleryfield-builtingalleryimage-crop"  # Can be omitted
    disable_server_side_crop = False

    def create_instance_from_form(self, form):
        self.object = form.save(commit=False)
        self.object.creator = self.request.user
        self.object.save()


class BuiltInImageListView(ImageListView):
    target_model = "galleryfield.BuiltInGalleryImage"
    crop_url_name = "galleryfield-builtingalleryimage-crop"  # Can be omitted
    disable_server_side_crop = False

    def get_queryset(self):
        queryset = super().get_queryset()
        if not self.request.user.is_superuser:
            queryset = queryset.filter(creator=self.request.user)
        return queryset


class BuiltInImageCropView(ImageCropView):
    target_model = "galleryfield.BuiltInGalleryImage"
    crop_url_name = "galleryfield-builtingalleryimage-crop"  # Can be commented
    disable_server_side_crop = False

    def get_object(self, queryset=None):
        obj = super().get_object(queryset=None)
        if not self.request.user.is_superuser and obj.creator != self.request.user:
            raise PermissionDenied("May not crop other user's image")
        return obj

    def create_cropped_instance_from_form(self, form):
        # we don't need to set self.object.creator here because
        # it's copied from the original instance.
        self.object = form.save()



import os

from django.test import TestCase

from demo_custom.models import CustomImage
from tests import factories
from tests.mixins import UserCreateMixin
from tests.utils import remove_upload_directory


class CustomDemoTest(UserCreateMixin, TestCase):
    @classmethod
    def setUpTestData(cls):  # noqa
        super().setUpTestData()
        cls.user = cls.create_user()

    def setUp(self) -> None:
        factories.UserFactory.reset_sequence()
        factories.CustomImageFactory.reset_sequence()
        factories.CustomGalleryFactory.reset_sequence()
        remove_upload_directory()

    # {{{ test remove image (instance) when removed from gallery in CustomGallery

    def test_update_gallery_with_image_removed(self):
        gallery = factories.CustomGalleryFactory.create(
            number_of_images=5, creator=self.user, user=self.user)

        first_image = CustomImage.objects.first()
        first_image_path = first_image.photo.path
        self.assertTrue(os.path.isfile(first_image_path))

        # Remove the first image
        gallery.images = gallery.images[1:]
        gallery.save()

        self.assertEqual(CustomImage.objects.count(), 4)
        self.assertFalse(os.path.isfile(first_image_path))

    def test_update_gallery_with_image_instance_not_removed(self):
        gallery = factories.CustomGalleryFactory.create(
            number_of_images=5, creator=self.user, user=self.user
        )

        first_image = CustomImage.objects.first()
        first_image_path = first_image.photo.path
        self.assertTrue(os.path.isfile(first_image_path))

        # another gallery with first image as member
        factories.CustomGalleryFactory.create(
            photos=[first_image], creator=self.user
        )

        # Remove the first image
        gallery.images = gallery.images[1:]
        gallery.save()

        self.assertEqual(CustomImage.objects.count(), 5)
        self.assertTrue(os.path.isfile(first_image_path))

    def test_update_gallery_with_image_instance_removed_but_file_persist(self):
        gallery = factories.CustomGalleryFactory.create(
            number_of_images=5, creator=self.user, user=self.user
        )

        first_image = CustomImage.objects.first()
        cloned_image = CustomImage.objects.first()
        cloned_image.pk = None
        cloned_image.save()

        first_image_path = first_image.photo.path
        self.assertTrue(os.path.isfile(first_image_path))
        self.assertEqual(first_image_path, cloned_image.photo.path)

        # Remove the first image
        gallery.images = gallery.images[1:]
        gallery.save()

        self.assertEqual(CustomImage.objects.count(), 5)
        self.assertTrue(os.path.isfile(first_image_path))

    # }}}



from django import forms
from django.contrib import admin

from demo_custom.models import CustomDemoGallery, CustomImage
from galleryfield.mixins import GalleryFormMediaMixin


class CustomDemoGalleryAdminForm(GalleryFormMediaMixin, forms.ModelForm):
    class Meta:
        model = CustomDemoGallery
        exclude = ()


class CustomDemoGalleryAdmin(admin.ModelAdmin):
    form = CustomDemoGalleryAdminForm


admin.site.register(CustomDemoGallery, CustomDemoGalleryAdmin)
admin.site.register(CustomImage)



from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase
from django.urls import NoReverseMatch, reverse

from galleryfield.utils import get_url_from_str


class GetUrlFromStrTest(SimpleTestCase):
    # Test galleryfield.utils.get_url_from_str

    def test_None(self):
        self.assertIsNone(get_url_from_str(None))
        self.assertIsNone(get_url_from_str(""))

    def test_valid_url_name(self):
        self.assertEqual(
            get_url_from_str("galleryfield-builtingalleryimage-upload"),
            reverse("galleryfield-builtingalleryimage-upload")
        )

    def test_invalid_url_name(self):
        with self.assertRaises(NoReverseMatch):
            # We are evaluating the lazy_result when compare
            self.assertEqual(get_url_from_str("non-exist-url-name"), "")
        with self.assertRaises(ImproperlyConfigured) as cm:
            get_url_from_str("non-exist-url-name", require_urlconf_ready=True)
        self.assertIn("is not a valid view function or pattern name",
                      cm.exception.args[0])

    def test_valid_url(self):
        url = reverse("gallery-update", kwargs={"pk": 1})
        self.assertEqual(
            get_url_from_str(url),
            url
        )

    def test_invalid_url(self):
        with self.assertRaises(NoReverseMatch):
            # We are evaluating the lazy_result when compare
            self.assertEqual(get_url_from_str("/foo/bar"), "")
        with self.assertRaises(ImproperlyConfigured) as cm:
            get_url_from_str("/foo/bar", require_urlconf_ready=True)
        self.assertIn("is not a valid view function or pattern name",
                      cm.exception.args[0])



from django.apps import AppConfig


class DemoCustomConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'demo_custom'

    def ready(self):
        import demo_custom.receivers  # noqa



from django.test import SimpleTestCase
from django.test.utils import override_settings


class CheckSettingsBase(SimpleTestCase):
    @property
    def func(self):
        from galleryfield.checks import check_settings
        return check_settings

    @property
    def msg_id_prefix(self):
        raise NotImplementedError()

    def assertCheckMessages(self,  # noqa
                            expected_ids=None, expected_msgs=None, length=None,
                            filter_message_id_prefixes=None, ignore_order=False):
        """
        Check the run check result of the setting item of the testcase instance
        :param expected_ids: Optional, list of expected message id,
        default to None
        :param expected_msgs: Optional, list of expected message string,
        default to None
        :param length: Optional, length of expected check message,
        default to None
        :param filter_message_id_prefixes: a list or tuple of message id prefix,
        to restrict the
         run check result to be within the iterable.
        """
        if not filter_message_id_prefixes:
            filter_message_id_prefixes = self.msg_id_prefix
            if isinstance(filter_message_id_prefixes, str):
                filter_message_id_prefixes = [filter_message_id_prefixes]
            assert isinstance(filter_message_id_prefixes, (list, tuple))

        if expected_ids is None and expected_msgs is None and length is None:
            raise RuntimeError("At least one parameter should be specified "
                               "to make the assertion")

        result = self.func(None)

        def is_id_in_filter(id, filter):
            prefix = id.split(".")[0]
            return prefix in filter

        try:
            result_ids, result_msgs = (
                list(zip(*[(r.id, r.msg) for r in result
                      if is_id_in_filter(r.id, filter_message_id_prefixes)])))

            if expected_ids is not None:
                assert isinstance(expected_ids, (list, tuple))
                if ignore_order:
                    result_ids = tuple(sorted(list(result_ids)))
                    expected_ids = sorted(list(expected_ids))
                self.assertEqual(result_ids, tuple(expected_ids))

            if expected_msgs is not None:
                assert isinstance(expected_msgs, (list, tuple))
                if ignore_order:
                    result_msgs = tuple(sorted(list(result_msgs)))
                    expected_msgs = sorted(list(expected_msgs))
                self.assertEqual(result_msgs, tuple(expected_msgs))

            if length is not None:
                self.assertEqual(len(expected_ids), len(result_ids))
        except ValueError as e:
            if "values to unpack" in str(e):
                if expected_ids or expected_msgs or length:
                    self.fail("Check message unexpectedly found to be empty")
            else:
                raise


class CheckConfigs(CheckSettingsBase):
    msg_id_prefix = "django-galleryfield"
    VALID_CONF_None = None
    VALID_CONF_Dict = {}
    INVALID_CONF_EMPTY_LIST = []
    INVALID_CONF_SPACES = " "

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_None)
    def test_valid_config1(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_Dict)
    def test_valid_config2(self):
        self.assertCheckMessages([])

    @override_settings(INSTALLED_APPS=['galleryfield', 'demo'])
    def test_invalid_config1(self):
        self.assertCheckMessages(["django-galleryfield.E001"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_EMPTY_LIST)
    def test_invalid_config2(self):
        self.assertCheckMessages(["django-galleryfield.E002"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_SPACES)
    def test_invalid_config3(self):
        self.assertCheckMessages(["django-galleryfield.E002"])


"""
DJANGO_GALLERY_FIELD_CONFIG = {
    "bootstrap_version": 3,
    "assets": {
        "jquery": 'vendor/jquery/dist/js/jquery.min.js'
        "bootstrap_js": 'vendor/bootstrap/dist/js/bootstrap.min.js',
        "bootstrap_css": "vendor/bootstrap/dist/css/bootstrap.min.css",
        "extra_js": [],
        "extra_css": [],
    },
    "thumbnails": {
        "size": 120,
        "quality": 80
    },
    jquery_file_upload_ui_options: {},
    "jquery_file_upload_ui_sortable_options": {}
    "widget_hidden_input_css_class": "django-galleryfield",
    "prompt_alert_if_changed_on_window_reload": True,
}
"""


class CheckBootstrapVersion(CheckSettingsBase):
    msg_id_prefix = "django-galleryfield-bootstrap_version"

    VALID_CONF_not_configured = {}

    VALID_CONF_None = {"bootstrap_version": None}

    VALID_CONF_3 = {"bootstrap_version": 3}

    VALID_CONF_3_str = {"bootstrap_version": "3"}

    VALID_CONF_5 = {"bootstrap_version": 5}

    INVALID_CONF_not_int = {"bootstrap_version": 3.4}

    INVALID_CONF_not_number = {"bootstrap_version": "3.4.1"}

    INVALID_CONF_2 = {"bootstrap_version": 2}

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_not_configured)
    def test_valid_config1(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_None)
    def test_valid_config2(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_3)
    def test_valid_config3(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_3_str)
    def test_valid_config4(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_5)
    def test_valid_config6(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_not_int)
    def test_invalid_config1(self):
        self.assertCheckMessages(['django-galleryfield-bootstrap_version.E001'])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_not_number)
    def test_invalid_config2(self):
        self.assertCheckMessages(['django-galleryfield-bootstrap_version.E001'])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_2)
    def test_invalid_config3(self):
        self.assertCheckMessages(['django-galleryfield-bootstrap_version.E002'])


class CheckAssets(CheckSettingsBase):
    msg_id_prefix = "django-galleryfield-assets"

    VALID_CONF_not_configured = {}

    VALID_CONF_None = {"assets": None}

    VALID_CONF_Dict = {"assets": dict()}

    VALID_CONF1 = {
        "assets": {
            "extra_js": ["some/js"],
            "extra_css": ["some/css"],
        },
    }

    VALID_CONF2 = {
        "assets": {
            "jquery": "foo",
            "bootstrap_css": "bar",
            "bootstrap_js": "bar",
            "extra_js": ["some/js"],
            "extra_css": ["some/css"],
        },
    }

    VALID_CONF_EXTRA_None = {
        "assets": {
            "extra_js": None,
            "extra_css": None,
        },
    }

    INVALID_CONF_NOT_Dict = {
        "assets": object
    }

    INVALID_CONF_EXTRA_JS = {
        "assets": {
            "extra_js": "extra_js",
        },
    }

    INVALID_CONF_UNKNOWN = {
        "assets": {
            # hyphen instead of underscore
            "bootstrap-js": "bar",
        },
    }

    INVALID_CONF_EXTRA_JS_ELE = {
        "assets": {
            "extra_js": ["extra_js", object, 1],
        },
    }

    INVALID_CONF_EXTRA_CSS = {
        "assets": {
            "extra_css": "extra_css",
        },
    }

    INVALID_CONF_EXTRA_CSS_ELE = {
        "assets": {
            "extra_css": ["extra_css", object, 2]
        },
    }

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_None)
    def test_valid_config1(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_Dict)
    def test_valid_config2(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF1)
    def test_valid_config3(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF2)
    def test_valid_config4(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_EXTRA_None)
    def test_valid_config5(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_not_configured)
    def test_valid_config6(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NOT_Dict)
    def test_invalid_config01(self):
        self.assertCheckMessages(["django-galleryfield-assets.E001"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_EXTRA_JS)
    def test_invalid_config02(self):
        self.assertCheckMessages(["django-galleryfield-assets.E002"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_EXTRA_CSS)
    def test_invalid_config03(self):
        self.assertCheckMessages(["django-galleryfield-assets.E004"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_EXTRA_JS_ELE)
    def test_invalid_config04(self):
        self.assertCheckMessages(["django-galleryfield-assets.E003",
                                  "django-galleryfield-assets.E003"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_EXTRA_CSS_ELE)
    def test_invalid_config05(self):
        self.assertCheckMessages(["django-galleryfield-assets.E005",
                                  "django-galleryfield-assets.E005"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_UNKNOWN)
    def test_invalid_config06(self):
        self.assertCheckMessages(["django-galleryfield-assets.E006"])


"""
    "thumbnails": {
        "size": 120,
        "quality": 80
    },
"""


class CheckThumbnails(CheckSettingsBase):
    msg_id_prefix = "django-galleryfield-thumbnails"

    VALID_CONF_None = {"thumbnails": None}

    VALID_CONF_Dict = {"thumbnails": dict()}

    VALID_CONF = {
        "thumbnails": {
            "size": 150,
            "quality": 95
        },
    }

    VALID_CONF_ALL_None = {
        "thumbnails": {
            "size": None,
            "quality": None
        },
    }

    VALID_CONF_int_tuple = {
        "thumbnails": {
            "size": (20, 80),
        },
    }

    VALID_CONF_int_list = {
        "thumbnails": {
            "size": (20, 80),
        },
    }

    VALID_CONF_str_tuple = {
        "thumbnails": {
            "size": ("20", "80"),
        },
    }

    VALID_CONF_str_list = {
        "thumbnails": {
            "size": ["20", "80"],
        },
    }

    VALID_CONF_x = {
        "thumbnails": {
            "size": "20x80",
        },
    }

    VALID_CONF_X = {
        "thumbnails": {
            "size": "20x80",
        },
    }

    VALID_CONF_CONTAINS_SPACES = {
        "thumbnails": {
            "size": " 20 x 80 ",
        },
    }

    VALID_CONF_CAN_CONVERT_TO_NUMBER = {
        "thumbnails": {
            "size": "40",
            "quality": "80"
        },
    }

    INVALID_CONF_NOT_Dict = {
        "thumbnails": object
    }

    INVALID_CONF_NOT_NUMBER = {
        "thumbnails": {
            "size": "one hundred",
            "quality": "ninety"
        },
    }

    INVALID_CONF_NEGATIVE = {
        "thumbnails": {
            "size": -10,
            "quality": -0.1
        },
    }

    INVALID_CONF_NEGATIVE_LIST = {
        "thumbnails": {
            "size": [-100, 100],
        },
    }

    INVALID_CONF_NEGATIVE_STR_LIST = {
        "thumbnails": {
            "size": ["-100", "100"],
        },
    }

    INVALID_CONF_NEGATIVE_TUPLE = {
        "thumbnails": {
            "size": ("-100", "100"),
        },
    }

    INVALID_CONF_EMPT_STR = {
        "thumbnails": {
            "size": "       ",
        },
    }

    INVALID_CONF_CONTAINS_3_tuple = {
        "thumbnails": {
            "size": (100, 200, 300),
        },
    }

    INVALID_CONF_CONTAINS_2_x = {
        "thumbnails": {
            "size": "100x200x300",
        },
    }

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_None)
    def test_valid_config1(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_Dict)
    def test_valid_config2(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF)
    def test_valid_config3(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_ALL_None)
    def test_valid_config4(self):
        self.assertCheckMessages([])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_CAN_CONVERT_TO_NUMBER)
    def test_valid_config5(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_int_tuple)
    def test_valid_config6(self):
        self.assertCheckMessages([])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_str_list)
    def test_valid_config7(self):
        self.assertCheckMessages([])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_int_list)
    def test_valid_config8(self):
        self.assertCheckMessages([])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_x)
    def test_valid_config9(self):
        self.assertCheckMessages([])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_X)
    def test_valid_config10(self):
        self.assertCheckMessages([])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_CONTAINS_SPACES)
    def test_valid_config11(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NOT_Dict)
    def test_invalid_config1(self):
        self.assertCheckMessages(["django-galleryfield-thumbnails.E001"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NOT_NUMBER)
    def test_invalid_config2(self):
        self.assertCheckMessages([
            "django-galleryfield-thumbnails.E003",
            "django-galleryfield-thumbnails.E004"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NEGATIVE)
    def test_invalid_config3(self):
        self.assertCheckMessages([
            "django-galleryfield-thumbnails.E003",
            "django-galleryfield-thumbnails.E005"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NEGATIVE_LIST)
    def test_invalid_config4(self):
        self.assertCheckMessages([
            "django-galleryfield-thumbnails.E003"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NEGATIVE_TUPLE)
    def test_invalid_config5(self):
        self.assertCheckMessages([
            "django-galleryfield-thumbnails.E003"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NEGATIVE_STR_LIST)
    def test_invalid_config6(self):
        self.assertCheckMessages([
            "django-galleryfield-thumbnails.E003"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_EMPT_STR)
    def test_invalid_config7(self):
        self.assertCheckMessages([
            "django-galleryfield-thumbnails.E002"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_CONTAINS_3_tuple)
    def test_invalid_config8(self):
        self.assertCheckMessages([
            "django-galleryfield-thumbnails.E003"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_CONTAINS_2_x)
    def test_invalid_config9(self):
        self.assertCheckMessages([
            "django-galleryfield-thumbnails.E003"])


"""
    jquery_file_upload_ui_options: {},
"""

MAX_NUMBER_OF_FILES = "maxNumberOfFiles"
SINGLE_FILE_UPLOADS = "singleFileUploads"
PREVIEW_MAX_WIDTH = "previewMaxWidth"
PREVIEW_MAX_HEIGHT = "previewMaxHeight"


class CheckJqueryFileUploadUiOptions(CheckSettingsBase):
    msg_id_prefix = "django-galleryfield-jquery_file_upload_ui_options"

    VALID_CONF_None = {"jquery_file_upload_ui_options": None}

    VALID_CONF_Dict = {"jquery_file_upload_ui_options": dict()}

    VALID_CONF = {
        "jquery_file_upload_ui_options": {
            "autoUpload": False
        }
    }

    WARN_CONF_MAX_NUMBER_OF_FILES = {
        "jquery_file_upload_ui_options": {
            "maxNumberOfFiles": 20
        }
    }

    WARN_CONF_SINGLE_FILE_UPLOADS_FALSE = {
        "jquery_file_upload_ui_options": {
            "singleFileUploads": False
        }
    }

    WARN_CONF_SINGLE_FILE_UPLOADS_FALSE_STRING = {
        "jquery_file_upload_ui_options": {
            "singleFileUploads": "false"
        }
    }

    NO_WARN_CONF_SINGLE_FILE_UPLOADS_TRUE = {
        "jquery_file_upload_ui_options": {
            "singleFileUploads": True
        }
    }

    WARN_CONF_PREVIEW_MAX_WIDTH = {
        "jquery_file_upload_ui_options": {
            "previewMaxWidth": 120
        }
    }

    WARN_CONF_PREVIEW_MAX_HEIGHT = {
        "jquery_file_upload_ui_options": {
            "previewMaxHeight": 180
        }
    }

    INVALID_CONF_NOT_Dict = {"jquery_file_upload_ui_options": object}

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_None)
    def test_valid_config1(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_Dict)
    def test_valid_config2(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF)
    def test_valid_config3(self):
        self.assertCheckMessages([])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=NO_WARN_CONF_SINGLE_FILE_UPLOADS_TRUE)
    def test_valid_config4(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NOT_Dict)
    def test_item_not_dict(self):
        self.assertCheckMessages(
            ["django-galleryfield-jquery_file_upload_ui_options.E001"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=WARN_CONF_MAX_NUMBER_OF_FILES)
    def test_item_warned_max_number_of_files(self):
        self.assertCheckMessages(
            ["django-galleryfield-jquery_file_upload_ui_options.W001"])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=WARN_CONF_SINGLE_FILE_UPLOADS_FALSE)
    def test_warn_single_upload_false_1(self):
        self.assertCheckMessages(
            ["django-galleryfield-jquery_file_upload_ui_options.W002"])

    @override_settings(
        DJANGO_GALLERY_FIELD_CONFIG=WARN_CONF_SINGLE_FILE_UPLOADS_FALSE_STRING)
    def test_warn_single_upload_false_2(self):
        self.assertCheckMessages(
            ["django-galleryfield-jquery_file_upload_ui_options.W002"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=WARN_CONF_PREVIEW_MAX_HEIGHT)
    def test_warn_preview_max_height(self):
        self.assertCheckMessages(
            ["django-galleryfield-jquery_file_upload_ui_options.W003"])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=WARN_CONF_PREVIEW_MAX_WIDTH)
    def test_warn_preview_max_width(self):
        self.assertCheckMessages(
            ["django-galleryfield-jquery_file_upload_ui_options.W003"])


"""
    jquery_file_upload_ui_sortable_options: {},
"""


class CheckJqueryFileUploadUiSortableOptions(CheckSettingsBase):
    msg_id_prefix = "django-galleryfield-jquery_file_upload_ui_sortable_options"

    VALID_CONF_None = {"jquery_file_upload_ui_sortable_options": None}

    VALID_CONF_Dict = {"jquery_file_upload_ui_sortable_options": dict()}

    INVALID_CONF_NOT_Dict = {"jquery_file_upload_ui_sortable_options": object}

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_None)
    def test_valid_config1(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_Dict)
    def test_valid_config2(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NOT_Dict)
    def test_item_not_dict(self):
        self.assertCheckMessages(
            ["django-galleryfield-jquery_file_upload_ui_sortable_options.E001"])


"""
    "widget_hidden_input_css_class": "django-galleryfield",
"""


class CheckMultifieldCssClassBasename(CheckSettingsBase):
    msg_id_prefix = "django-galleryfield-widget_hidden_input_css_class"

    VALID_CONF_None = {"widget_hidden_input_css_class": None}

    VALID_CONF_str = {"widget_hidden_input_css_class": "some-class-name"}

    INVALID_CONF_NOT_str = {"widget_hidden_input_css_class": object}

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_None)
    def test_valid_config1(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_str)
    def test_valid_config2(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NOT_str)
    def test_invalid_config1(self):
        self.assertCheckMessages([
            "django-galleryfield-widget_hidden_input_css_class.E001"])


"""
    "prompt_alert_if_changed_on_window_reload": True,
"""


class CheckPromptAlert(CheckSettingsBase):
    msg_id_prefix = "django-galleryfield-prompt_alert_if_changed_on_window_reload"

    VALID_CONF_None = {"prompt_alert_if_changed_on_window_reload": None}

    VALID_CONF_bool = {"prompt_alert_if_changed_on_window_reload": False}

    INVALID_CONF_NOT_bool = {"prompt_alert_if_changed_on_window_reload": object}

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_None)
    def test_valid_config1(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=VALID_CONF_bool)
    def test_valid_config2(self):
        self.assertCheckMessages([])

    @override_settings(DJANGO_GALLERY_FIELD_CONFIG=INVALID_CONF_NOT_bool)
    def test_invalid_config1(self):
        self.assertCheckMessages([
            "django-galleryfield-prompt_alert_if_changed_on_window_reload.E001"])



from django.db import transaction
from django.db.models import FileField
from django.db.models.signals import post_delete, post_save
from django.dispatch.dispatcher import receiver

from demo_custom.models import CustomDemoGallery, CustomImage


@receiver(post_save, sender=CustomDemoGallery)
@transaction.atomic
def refresh_gallery_image_set(
        sender, created, instance: CustomDemoGallery, **kwargs):
    related_images = instance.related_images

    # CustomImage objects which were in the gallery but now were removed
    images_to_delete = list(related_images.exclude(pk__in=list(instance.images)))

    # Clear and update the m2m relationship to the latest value
    related_images.clear()
    instance.related_images.set(instance.images.objects.all())

    # Delete CustomImage objects which don't belong to other gallery instance
    for image in images_to_delete:
        if not image.gallery.count():
            image.delete()


# {{{ Delete files when delete model instance
# https://cmljnelson.blog/2020/06/22/delete-files-when-deleting-models-in-django/


def delete_file_if_unused(model, instance, field, instance_file_field):
    """ Only delete the file if no other instances of that model are using it"""
    dynamic_field = dict()
    dynamic_field[field.name] = instance_file_field.name
    other_refs_exist = model.objects.filter(
        **dynamic_field).exclude(pk=instance.pk).exists()
    if not other_refs_exist:
        instance_file_field.delete(False)


@receiver(post_delete, sender=CustomImage)
def delete_files_when_row_deleted_from_db(sender, instance, **kwargs):
    """ Whenever ANY model is deleted, if it has a file field on it,
    delete the associated file too"""

    for field in sender._meta.concrete_fields:
        if isinstance(field, FileField):
            instance_file_field: FileField = getattr(instance, field.name)
            delete_file_if_unused(sender, instance, field, instance_file_field)

# }}}



from time import sleep
from urllib.parse import urljoin

from django.contrib.staticfiles.finders import find
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.test.utils import override_settings
from django.urls import reverse
from selenium import webdriver
from selenium.webdriver.common.by import By

from demo.models import DemoGallery
from demo_custom.models import CustomDemoGallery, CustomImage
from galleryfield.models import BuiltInGalleryImage
from tests.mixins import UserCreateMixin
from tests.utils import remove_upload_directory, test_media_root

CHROMIUM = "chromium"
FIREFOX = "firefox"
SELENIUM_BROWSER = CHROMIUM


class SeleniumTestMixin(UserCreateMixin):

    @property
    def gallery_create_url_name(self):
        raise NotImplementedError

    @property
    def gallery_detail_url_name(self):
        raise NotImplementedError

    @property
    def image_model(self):
        raise NotImplementedError

    @property
    def gallery_model(self):
        raise NotImplementedError

    @classmethod
    def setUpClass(cls):
        if SELENIUM_BROWSER == CHROMIUM:
            options = webdriver.ChromeOptions()
            options.add_argument("--headless")
            options.add_argument("--disable-gpu")
            cls.selenium = webdriver.Chrome(options=options)
        elif SELENIUM_BROWSER == FIREFOX:
            cls.selenium = webdriver.Firefox()
        super().setUpClass()

    def setUp(self):
        super().setUp()
        remove_upload_directory()
        self.superuser = self.create_superuser()

    @classmethod
    def tearDownClass(cls):
        cls.selenium.quit()
        super().tearDownClass()

    def get_full_url(self, url):
        return urljoin(self.live_server_url, url)

    @property
    def admin_url(self):
        return self.get_full_url("/admin/")

    @property
    def new_gallery_url(self):
        return self.get_full_url(reverse(self.gallery_create_url_name))

    def get_gallery_detail_url(self, pk):
        return self.get_full_url(reverse(
            self.gallery_detail_url_name, kwargs={"pk": pk}))

    def _login_to_admin(self):
        # We do this because we need to login
        self.selenium.get(self.admin_url)
        username_input = self.selenium.find_element(by=By.NAME, value="username")
        username_input.send_keys(self.superuser.username)
        password_input = self.selenium.find_element(by=By.NAME, value="password")
        password_input.send_keys("secret")
        self.selenium.find_element(
            by=By.XPATH, value='//input[@value="Log in"]').click()
        sleep(1)

    def _go_to_new_gallery(self):
        self.selenium.get(self.new_gallery_url)

    def _go_to_gallery_detail(self, pk):
        self.selenium.get(self.get_gallery_detail_url(pk))
        sleep(1)

    def _assert_widget_loaded(self):
        sleep(1)
        self.selenium.find_element(by=By.ID, value="images-files")

    def _add_and_upload_image(self, static_file_path):
        input_ele = self.selenium.find_element(by=By.ID, value="images-files")
        input_ele.send_keys(self._get_upload_image(static_file_path))
        sleep(0.5)
        self.selenium.find_element(by=By.CLASS_NAME, value="start").click()
        sleep(1)

    def _get_upload_image(self, static_file_path):
        return find(static_file_path)

    def _submit_page(self):
        submit_ele = self.selenium.find_element(by=By.ID, value="submit-id-submit")
        submit_ele.click()
        sleep(0.5)

    def _get_first_button_bar_delete_button(self):
        return self.selenium.find_element(by=By.CSS_SELECTOR,
                                          value="a.delete.fileinput-button")

    def _delete_one_image(self):
        delete_button = self._get_first_button_bar_delete_button()
        delete_button.click()
        sleep(0.5)

    def test_upload_image(self):
        self.assertEqual(self.gallery_model.objects.count(), 0)
        self._login_to_admin()
        self._go_to_new_gallery()
        self._assert_widget_loaded()
        self._add_and_upload_image("demo/screen_upload.png")
        self.assertEqual(self.image_model.objects.count(), 1)
        self._submit_page()
        self._add_and_upload_image("demo/screen_crop.png")
        self.assertEqual(self.image_model.objects.count(), 2)
        self._submit_page()
        self.assertEqual(self.gallery_model.objects.count(), 1)
        self.assertEqual(
            self.gallery_model.objects.first().images.objects.count(), 2)
        self._go_to_gallery_detail(self.gallery_model.objects.first().pk)

    def test_delete_image(self):
        self.assertEqual(self.gallery_model.objects.count(), 0)
        self._login_to_admin()
        self._go_to_new_gallery()
        self._assert_widget_loaded()
        self._add_and_upload_image("demo/screen_upload.png")
        self._add_and_upload_image("demo/screen_crop.png")

        self._delete_one_image()
        self._submit_page()
        self.assertEqual(self.gallery_model.objects.count(), 1)
        self.assertEqual(
            self.gallery_model.objects.first().images.objects.count(), 1)

        self._delete_one_image()
        self._submit_page()
        self.assertEqual(self.gallery_model.objects.count(), 1)
        # Because this is the last image, delete is not allowed
        # because this field is required
        self.assertEqual(
            self.gallery_model.objects.first().images.objects.count(), 1)

    def test_delete_one_before_saving(self):
        self.assertEqual(self.gallery_model.objects.count(), 0)
        self._login_to_admin()
        self._go_to_new_gallery()
        self._assert_widget_loaded()
        self._add_and_upload_image("demo/screen_upload.png")
        self._add_and_upload_image("demo/screen_crop.png")

        # Delete the first uploaded image
        self.assertEqual(self.image_model.objects.count(), 2)
        self.image_model.objects.first().delete()
        self.assertEqual(self.image_model.objects.count(), 1)

        self._submit_page()
        self.assertEqual(self.gallery_model.objects.count(), 1)
        self.assertEqual(
            self.gallery_model.objects.first().images.objects.count(), 1)


@override_settings(MEDIA_ROOT=test_media_root)
class TestDemoApp(SeleniumTestMixin, StaticLiveServerTestCase):

    gallery_create_url_name = "gallery"

    gallery_detail_url_name = "gallery-detail"

    image_model = BuiltInGalleryImage

    gallery_model = DemoGallery


@override_settings(MEDIA_ROOT=test_media_root)
class TestCustomDemoApp(SeleniumTestMixin, StaticLiveServerTestCase):

    gallery_create_url_name = "custom-gallery"

    gallery_detail_url_name = "custom-gallery-detail"

    image_model = CustomImage

    gallery_model = CustomDemoGallery



from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.http.response import HttpResponseRedirect
from django.urls import reverse
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView

from demo_custom.forms import CustomGalleryForm
from demo_custom.models import CustomDemoGallery

# class GalleryEditPermissionTestMixin(UserPassesTestMixin):
#     raise_exception = True
#
#     def test_func(self):
#         user_id = int(self.kwargs["user_id"])
#         try:
#             user = get_object_or_404(get_user_model(), id=user_id)
#         except Exception:
#             return False
#         else:
#             if self.request.user.is_superuser:
#                 return True
#             if self.request.user == user:
#                 return True
#             return False


class CustomGalleryCreateView(LoginRequiredMixin, CreateView):
    form_class = CustomGalleryForm
    template_name = "form.html"
    extra_context = {
        "form_description": "Create a new custom gallery"}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        return context

    def form_valid(self, form):
        self.object = form.save(commit=False)
        self.object.creator = self.request.user
        self.object.save()
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        return reverse("custom-gallery-update", kwargs={"pk": self.object.pk})


class CustomGalleryUpdateView(UserPassesTestMixin, LoginRequiredMixin, UpdateView):
    model = CustomDemoGallery
    form_class = CustomGalleryForm
    template_name = "form.html"
    extra_context = {
        "form_description": "Update Custom Gallery"}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["detail_view_url"] = reverse(
            "custom-gallery-detail", kwargs={"pk": self.object.pk})
        return context

    def test_func(self):
        return (self.request.user == self.get_object().creator
                or self.request.user.is_superuser)


class CustomGalleryDetailView(DetailView):
    # Anonymous user can visit this.

    model = CustomDemoGallery
    extra_context = {
        "form_description": "Custom Gallery Detail"}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["update_view_url"] = reverse(
            "custom-gallery-update", kwargs={"pk": self.object.pk})
        return context



from django.contrib.auth.decorators import login_required
from django.urls import path, re_path

from demo_custom import image_views, views

urlpatterns = [
    # image handle views

    path(r'upload/',
         login_required(image_views.CustomImageCreateView.as_view()),
         name="demo_custom-customimage-upload"),
    path(r'fetch/',
         login_required((image_views.CustomImageListView.as_view())),
         name="demo_custom-customimage-fetch"),
    path(r'crop/<int:pk>',
         login_required(image_views.CustomImageCropView.as_view()),
         name="customimage-crop"),

    # gallery views
    path(r'', views.CustomGalleryCreateView.as_view(), name='custom-gallery'),
    path('update/<int:pk>',
         views.CustomGalleryUpdateView.as_view(), name='custom-gallery-update'),
    path('detail/<int:pk>',
         views.CustomGalleryDetailView.as_view(), name='custom-gallery-detail'),

    re_path(r"^user_images"
            + r"/(?P<user_id>\d+)"
            r"/(?P<image_id>\d+)"
            r"/(?P<file_name>[^/]+)$",
            image_views.image_download,
            name="image_download"),
]



# Generated by Django 3.2.13 on 2022-05-21 11:59

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import demo_custom.models
import galleryfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomDemoGallery',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('images', galleryfield.fields.GalleryField(blank=True, null=True, target_model='demo_custom.CustomImage', verbose_name='Photos')),
                ('creator', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='Creator')),
            ],
        ),
        migrations.CreateModel(
            name='CustomImage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('photo', models.ImageField(storage=demo_custom.models.UserImageStorage(), upload_to=demo_custom.models.user_img_path, verbose_name='Photo')),
                ('added_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('gallery', models.ManyToManyField(blank=True, related_name='related_images', to='demo_custom.CustomDemoGallery', verbose_name='Gallery belongs to')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='Owner')),
            ],
        ),
    ]



from django.conf import settings
from django.core.files.storage import FileSystemStorage
from django.db import models
from django.urls import reverse
from django.utils.deconstruct import deconstructible
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from galleryfield.fields import GalleryField


class CustomDemoGallery(models.Model):
    """
    The gallery model used in demo_custom.

    """
    images = GalleryField(
        verbose_name=_('Photos'), target_model='demo_custom.CustomImage',
        blank=True, null=True)
    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True,
        verbose_name=_('Creator'), on_delete=models.CASCADE)

    def get_absolute_url(self):
        return reverse('custom-gallery-update', kwargs={'pk': self.pk})


@deconstructible
class UserImageStorage(FileSystemStorage):
    def __init__(self):
        super().__init__(location=settings.SENDFILE_ROOT)


storage = UserImageStorage()


def user_img_path(instance, filename):
    return "user_images/user_{}/{}".format(instance.user_id, filename)


class CustomImage(models.Model):
    photo = models.ImageField(
        upload_to=user_img_path, storage=storage, verbose_name=_("Photo"))

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=False, blank=False,
        verbose_name=_('Owner'), on_delete=models.CASCADE)

    added_at = models.DateTimeField(default=now)

    # This is used to determine where the image is an orphan
    gallery = models.ManyToManyField(
        CustomDemoGallery, blank=True,
        verbose_name=_("Gallery belongs to"), related_name="related_images")

    def get_image_url(self):
        import os
        file_name = os.path.basename(self.photo.path)

        return reverse(
            "image_download", args=[
                self.user_id,
                self.pk,
                file_name
            ])

    def get_crop_url(self):
        return reverse(
            "customimage-crop", args=[
                self.pk
            ])

    def serialize_extra(self, request):
        from django.utils import formats

        return {
            "added_datetime": formats.date_format(
                self.added_at, "SHORT_DATETIME_FORMAT")}

    @classmethod
    def get_image_field(cls):
        return cls._meta.get_field("photo")



from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from django_sendfile import sendfile

from demo_custom.models import CustomImage
from galleryfield.image_views import (ImageCreateView, ImageCropView,
                                      ImageListView)


class CustomImageCreateView(LoginRequiredMixin, ImageCreateView):
    target_model = "demo_custom.CustomImage"
    disable_server_side_crop = False

    def create_instance_from_form(self, form):
        self.object = form.save(commit=False)
        self.object.user = self.request.user
        self.object.save()


class CustomImageListView(LoginRequiredMixin, ImageListView):
    target_model = "demo_custom.CustomImage"
    disable_server_side_crop = False

    def get_queryset(self):
        queryset = super().get_queryset()
        return queryset.filter(user=self.request.user)


class CustomImageCropView(LoginRequiredMixin, ImageCropView):
    target_model = "demo_custom.CustomImage"
    disable_server_side_crop = False

    def create_cropped_instance_from_form(self, form):
        self.object: CustomImage = form.save(commit=False)
        self.object.save()


# {{{ sendfile


@login_required
def image_download(request, **kwargs):
    user_id = kwargs["user_id"]  # noqa
    photo_id = kwargs["image_id"]
    file_name = kwargs["file_name"]  # noqa

    from demo_custom.models import CustomImage

    download_object = get_object_or_404(CustomImage, pk=photo_id)

    privilege = False

    if request.user.is_staff:
        privilege = True

    return _auth_download(request, download_object, privilege)


@login_required
def _auth_download(request, download_object, privilege=False):
    if request.user == download_object.user or privilege:
        return sendfile(request, download_object.photo.path)

    raise PermissionDenied(_("may not view other user's image"))

# }}}



from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit
from django import forms

from demo_custom.models import CustomDemoGallery


class CustomGalleryForm(forms.ModelForm):
    class Meta:
        model = CustomDemoGallery
        fields = ["images"]

        # Remove label
        labels = {'images': ""}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["images"].widget.download_template = (
            "demo_custom/custom_download_template.html")

        self.helper = FormHelper(self)
        self.helper.layout.append(
            Submit("Submit", "submit",
                   css_class="gallery-widget-submit-button"))
